package request;

import arithmetic.Operation;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

public class RequestHandler
{
    private Map<Key, Result> results;
    private Lock resultsLock;

    public RequestHandler()
    {
        results = new HashMap<>();
        resultsLock = new ReentrantLock();
    }   
    
    // ritorna il risultato appena e' pronto o dopo timeout millisec (con null)
    public Result syncRequest(Request req, long timeout)
    {
        return syncRetrive(asyncRequest(req), timeout);
    }
    
    // ritorna una chiave unica per recuperare il risultato
    public Key asyncRequest(Request req)
    {
        Key k = new Key(resultsLock); // TODO: se e' gia' stata fatta una richiesta uguale ritorna la stessa key 
        new Thread(new ArithmeticExecutor(req.getOperation(), k, results, resultsLock)).start();
        return k;
    }
    
    // ritorna appena il risultato e' pronto o dopo timeout millisec (con null)
    public Result syncRetrive(Key k, long timeout)
    {
        resultsLock.lock();
        try
        {
            while (!results.containsKey(k))
            {
                try
                {
                    boolean ok = k.getCondition().await(timeout, TimeUnit.MILLISECONDS);
                    if (!ok)
                    {
                        System.out.println("Ho aspettato troppo!");
                        return null; // ho aspettato timeout millisec: esco
                    }
                }
                catch (InterruptedException ie)
                {
                    ie.printStackTrace();
                }
            }
            return results.get(k);
        }
        finally
        {
            resultsLock.unlock();
        }
    }
    
    // ritorna null se il risultato non e' pronto
    public Result asyncRetrieve(Key k)
    {
        resultsLock.lock();
        try
        {
            if (results.containsKey(k))
            {
                return results.get(k);
            }
        }
        finally
        {
            resultsLock.unlock();
        }
        return null;
    }
    
    public static void main(String[] args)
    {
        String s1 = "3 4 5 * +"; // -> 23
        String s2 = " 3 4 + 5 *"; // -> 35
        String s3 = "3 4 5 + *"; // -> 27
        RequestHandler reqHandler = new RequestHandler();
        
        Operation op = Operation.fromString(s1);
        Request r1 = new Request(op);
        Request r2 = new Request(Operation.fromString(s2));
        Request r3 = new Request(Operation.fromString(s3));
        Request r4 = new Request(op); // stesso oggetto di r1
      
        Key k1 = reqHandler.asyncRequest(r1);
        Key k2 = reqHandler.asyncRequest(r2);
        Key k3 = reqHandler.asyncRequest(r3);
        
        System.out.println("Retrieve 2 asincrona: " + reqHandler.asyncRetrieve(k2));
        System.out.println("Retrieve 1 sincrona: " + reqHandler.syncRetrive(k1, 10000));
        System.out.println("Retrieve 3 sincrona: " + reqHandler.syncRetrive(k3, 10000));
        System.out.println("Request 4 sincrona: " + reqHandler.syncRequest(r4, 10000));
    }
}